//     :     :     :     :     :     :     :     :     :     :
//   --------:  ---------:  ---------:  ---------:  ----     :
// ..|      |*..|       |*..|       |*..|       |*..|  |*....:.
//   |      |*  |       |*  |       |*  |       |*  |  |*    :
//   ---  ---*  |  |-----*  |  |-----*  |  |-----*  |  |*    :
// ....|  |*.:..|  |--****..|  |-----*..|  |--****..|  |*....:.
//     |  |* :  |    |*  :  |  |--  |*  |    |*  :  |  |*    :
//     |  |* :  |  |--*  :  |  |*|  |*  |  |--*  :  |  |*    :
// ....|  |*.:..|  |-----:..|  ---  |*..|  |-----:..|  -----.:.
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
// ....----*.:..---------*..---------*..---------*..--------*:.
//     :**** :   *********   *********   *********   ********:
//     :     :     :     :     :     :     :     :     :     :
//
// Version 0.81
// Copyright (c) 2002-2007
// Martijn "Amarillion" van Iersel.
//
// This file is part of the Tegel tilemapping add-on for allegro
//
//    Tegel is free software; you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General 
//    Public License as published by the Free Software 
//    Foundation; either version 2.1 of the License, or
//    (at your option) any later version.
//
//    Tegel is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with Tegel; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
// official website:
// http://tegel.sourceforge.net/
//
// I hope you like this program. Please send 
// questions, comments, flames and feature 
// requests to amarillion@yahoo.com.
//
#ifndef MAPBOX_H
#define MAPBOX_H

#include "gridbox.h"
#include "tegel.h"
#include "data.h"
#include "undo.h"
#include <vector>
#include <assert.h>

enum { T_SELECT, T_BRUSH, T_FILL };

class CheckeredPattern
{
	private:
	BITMAP *checkered_pattern;
	BITMAP *checkered_pattern_flip;
	
	public:
	CheckeredPattern()
	{
		checkered_pattern = create_bitmap (8, 8);
		checkered_pattern_flip = create_bitmap (8, 8);
		clear_to_color (checkered_pattern, MAS::Color::black);
		clear_to_color (checkered_pattern_flip, MAS::Color::white);
		rectfill (checkered_pattern, 0, 0, 3, 3, MAS::Color::white);
		rectfill (checkered_pattern, 4, 4, 7, 7, MAS::Color::white);
		rectfill (checkered_pattern_flip, 0, 0, 3, 3, MAS::Color::black);
		rectfill (checkered_pattern_flip, 4, 4, 7, 7, MAS::Color::black);
	}
	
	~CheckeredPattern()
	{
		destroy_bitmap (checkered_pattern);
		destroy_bitmap (checkered_pattern_flip);
	}
	
	void checkeredRect (BITMAP *target, int x1, int y1, int x2, int y2, bool flip)
	{
		drawing_mode (DRAW_MODE_COPY_PATTERN, flip ? 
			checkered_pattern : checkered_pattern_flip, 0, 0);
		rect (target, x1, y1, x2, y2, 0);
		solid_mode();
	}
};

// handles all scrolling of the map
class ScrollWrapper : public DataWrapper
{
    private:
        int x, y; // current offset
        int xvis, yvis; // w, h of viewport
        int xtot, ytot; // w, h of map
        void CheckBounds();
    public:
        ScrollWrapper();
        ~ScrollWrapper();
        void SetVisible (int x, int y);
        void SetTotal (int x, int y);
        void SetPos (int _x, int _y);
        void Scroll (int dx, int dy);
		void ScrollToArea (int xx, int yy, int ww, int hh);
        int GetX() { return x; }
        int GetY() { return y; }
        int GetXVis() { return xvis; }
        int GetYVis() { return yvis; }
};

class MapWrapper : public DataWrapper
{
    friend class UndoBrush;
    friend class UndoTotal;
    private:
        TEG_MAP *map;
        bool changed;
		bool titled;
        UndoList undolist;
		int cursorx, cursory;
		std::string filename;
		// helper for cut & delete
		void BlockSet (int layer, int x, int y, int dl, int w, int h, int value);
    public:
		enum {EVT_MAP_CONTENT, EVT_MAP_TOTAL, EVT_CURSOR, EVT_FILENAME};
        MapWrapper();
        ~MapWrapper();
        void SetMap(TEG_MAP *newmap);
        TEG_MAP *GetMap() { return map; }
        int GetCursorX() { return cursorx; }
		int GetCursorY() { return cursory; }
		void SetCursor(int x, int y);
		void Set (int layer, int x, int y, int value);
		void SetFile (std::string _filename, bool _titled, bool _changed);
        void FloodFill (int layer, int x, int y, int value);
        void Resize (int newdl, int neww, int newh);
		void Paste (TEG_MAP *map, int l, int x, int y);
        bool HasChanged() { return changed; }
        bool IsTitled() { return titled; }
		void Undo();
		void Delete (int layer, int x, int y, int dl, int w, int h);
		void Cut (int layer, int x, int y, int dl, int w, int h);
		std::string GetFilename() { return filename; }
        std::string GetUndoDescription() { return undolist.GetDescription(); }
};

class LayerWrapper : public DataWrapper
{
    private:
        std::vector <bool> visible;
        int layernum;
        int active;
    public:
        LayerWrapper();
        ~LayerWrapper();
        int GetActiveLayer() { return active; }
        bool IsVisible (int layer);
        void InverseVisible (int layer);
        void Reset(); // call when new map
        int GetLayerNum () { return layernum; }
        void SetLayerNum (int _layernum); // call whenever layer num changed
        void SetActive (int _active);
};

class MapBox;

class MapBoxLayerListener : public DataListener
{
    private:
        MapBox *parent;
    public:
        MapBoxLayerListener (MapBox *p)
        {
            assert (p);
            parent = p;
        }
		virtual ~MapBoxLayerListener () {}
        virtual void changed (int code);
};

class MapBoxMapListener : public DataListener
{
    private:
        MapBox *parent;
    public:
        MapBoxMapListener (MapBox *p)
        {
            assert (p);
            parent = p;
        }
        virtual void changed (int code);
};

class MapBoxScrollListener : public DataListener
{
    private:
        MapBox *parent;
    public:
        MapBoxScrollListener (MapBox *p)
        {
            assert (p);
            parent = p;
        }
        virtual void changed (int code);
		virtual ~MapBoxScrollListener() {}
};

class MapBox : public GridBox
{
	private:
		int tool, currentTile, currentTileR;
		//TEG_MAP *map;
		
		MapWrapper *mapWrap;
		LayerWrapper *layerWrap;
		ScrollWrapper *scrollWrap;
		
		MapBoxLayerListener layerListener;
		MapBoxMapListener mapListener;
		MapBoxScrollListener scrollListener;
		CheckeredPattern checkeredPattern;
		bool blockScrollEvent; // used to block events generated by self.
	
		int frame;
		int timerId;
	
		int cursorTimerId;
		bool cursorBlink;
		bool clientfocus; // focus of client area! Added to prevent drawing tiles when the client really doesn't have the focus
		int scrollRepeatTimerId;
		
		enum SelectStates {ssNormal, ssSelecting};
		
		int selectx;
		int selecty;
		int selectw;
		int selecth;
		int selectl;
		int selectdl;
		int select_startx;
		int select_starty;
		SelectStates selectState;

		TEG_MAP *clipBoard; // map area in clipboard for copy / paste actions
		
		void UpdateSelection ();
		void ScrollToCursor();
	protected:
		virtual void MsgTimer (int t);
		virtual void MsgMousemove(const MAS::Point &d);
		virtual void DrawClientArea (MAS::Bitmap &dest, int xofst, int yofst);
		virtual void DrawCell (MAS::Bitmap &dest, int col, int row, int x, int y);        
		virtual void MsgScroll(); 
		virtual void MsgLPress();
		virtual void MsgLRelease();
		virtual void MsgRPress();
		virtual bool MsgXChar(int c);
		virtual void MsgGotMouse();
		virtual void MsgLostMouse();
	public:   
		MapBox (MapWrapper *_mapWrap, 
		    LayerWrapper *_layerWrap, ScrollWrapper *_scrollWrap);
		virtual ~MapBox();
		//setters and getters
		void setTool (int value) { tool = value; } 
		void setCurrentTile (int value) { currentTile = value; }
		void setCurrentTileR (int value) { currentTileR = value; }
		void LayersChanged (int code);
		void ScrollChanged (int code);
		void MapChanged (int code);
		void Copy (); // copy current selection to clipboard
		void Paste (); // paste clipboard as floating selection
		void Cut ();
		void Delete ();
};

#endif
