//     :     :     :     :     :     :     :     :     :     :
//   --------:  ---------:  ---------:  ---------:  ----     :
// ..|      |*..|       |*..|       |*..|       |*..|  |*....:.
//   |      |*  |       |*  |       |*  |       |*  |  |*    :
//   ---  ---*  |  |-----*  |  |-----*  |  |-----*  |  |*    :
// ....|  |*.:..|  |--****..|  |-----*..|  |--****..|  |*....:.
//     |  |* :  |    |*  :  |  |--  |*  |    |*  :  |  |*    :
//     |  |* :  |  |--*  :  |  |*|  |*  |  |--*  :  |  |*    :
// ....|  |*.:..|  |-----:..|  ---  |*..|  |-----:..|  -----.:.
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
// ....----*.:..---------*..---------*..---------*..--------*:.
//     :**** :   *********   *********   *********   ********:
//     :     :     :     :     :     :     :     :     :     :
//
// Version 0.81
// Copyright (c) 2002-2007
// Martijn "Amarillion" van Iersel.
//
// This file is part of the Tegel tilemapping add-on for allegro
//
//    Tegel is free software; you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General 
//    Public License as published by the Free Software 
//    Foundation; either version 2.1 of the License, or
//    (at your option) any later version.
//
//    Tegel is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with Tegel; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
// official website:
// http://tegel.sourceforge.net/
//
// I hope you like this program. Please send 
// questions, comments, flames and feature 
// requests to amarillion@yahoo.com.
//
#include "mapbox.h"
#include "tegel.h"
#include "flood.h"
#include "undo.h"
#include <assert.h>

using namespace MAS;

ScrollWrapper::ScrollWrapper() : DataWrapper()
{
    x = 0;
    y = 0;
    xvis = 0;
    yvis = 0;
    xtot = 0;
    ytot = 0;
}

ScrollWrapper::~ScrollWrapper()
{
}

void ScrollWrapper::CheckBounds()
{
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    if (x + xvis >= xtot) x = xtot - xvis;
    if (y + yvis >= ytot) y = ytot - yvis;
}

void ScrollWrapper::SetVisible (int x, int y)
{
    if (x != xvis || y != yvis)
    {
        xvis = x;
        yvis = y;
        CheckBounds();
        FireEvent (0);
    }
}
    
void ScrollWrapper::SetTotal (int x, int y)
{
    if (x != xtot || y != ytot)
    {        
        xtot = x;
        ytot = y;
        CheckBounds();
        FireEvent (0);
    }
}

void ScrollWrapper::SetPos (int _x, int _y)
{
    int oldx = x;
    int oldy = y;
    x = _x;
    y = _y;
    CheckBounds();
    if (x != oldx || y != oldy)
    {
        FireEvent (0);
    }
}

void ScrollWrapper::Scroll (int dx, int dy)
{
    int oldx = x;
    int oldy = y;
    x += dx;
    y += dy;    
    CheckBounds();
    if (x != oldx || y != oldy)
    {
        FireEvent (0);
    }
}

void ScrollWrapper::ScrollToArea (int xx, int yy, int ww, int hh)
{
	int oldx = x;
	int oldy = y;
	if (xx + ww < x || xx > x + xvis)
	{
		x = xx - (xvis / 2);
	}
	if (yy + hh < y || yy > y + yvis)
	{
		y = yy - (yvis / 2);
	}
	CheckBounds();
	if (x != oldx || y != oldy)
	{
		FireEvent (0);
	}
}

MapWrapper::MapWrapper() : DataWrapper(), undolist()
{
    map = NULL;
    changed = false;
	titled = false;
	filename = "";
	cursorx = 0;
	cursory = 0;
}

MapWrapper::~MapWrapper()
{
    if (map) teg_destroymap (map);
}

void MapWrapper::SetFile (std::string _filename, bool _titled, bool _changed)
{
	filename = _filename;
	titled = _titled;
	changed = _changed;
	FireEvent (EVT_FILENAME);
}
        
void MapWrapper::Undo ()
{
    undolist.Undo(this);
	changed = true;
    FireEvent (EVT_MAP_TOTAL);
    FireEvent (EVT_CURSOR);
}

void MapWrapper::SetMap(TEG_MAP *newmap)
{
    if (map) teg_destroymap (map);
    map = newmap;
    undolist.Clear();
	changed = false;
	cursorx = 0;
	cursory = 0;
    FireEvent (EVT_MAP_TOTAL);
    FireEvent (EVT_CURSOR);
}

void MapWrapper::Set (int layer, int x, int y, int value)
{
    undolist.Add (new UndoBrush (layer, x, y, teg_mapget (map, layer, x, y)));
    teg_mapput (map, layer, x, y, value);
    changed = true;
    FireEvent (EVT_MAP_CONTENT);
}

// helper for cut & delete
void MapWrapper::BlockSet (int layer, int x, int y, int dl, int w, int h, int value)
{
	for (int il = layer; il < layer + dl; ++il)
		for (int iy = y; iy < y + h; ++iy)
			for (int ix = x; ix < x + w; ++ix)
				teg_mapput (map, il, ix, iy, value);
}

void MapWrapper::Cut (int layer, int x, int y, int dl, int w, int h)
{
	undolist.Add (new UndoTotal (map, "cut", x, y));
	BlockSet (layer, x, y, dl, w, h, -1);
	changed = true;
	FireEvent (EVT_MAP_CONTENT);
}

void MapWrapper::Delete (int layer, int x, int y, int dl, int w, int h)
{
	undolist.Add (new UndoTotal (map, "delete", x, y));
	BlockSet (layer, x, y, dl, w, h, -1);
	changed = true;
	FireEvent (EVT_MAP_CONTENT);
	
}

void MapWrapper::FloodFill (int layer, int x, int y, int value)
{
    undolist.Add (new UndoTotal (map, "floodfill", x, y));
    teg_floodfill (map, layer, x, y, value);
    changed = true;
    FireEvent (EVT_MAP_CONTENT);
}
    
void MapWrapper::Resize (int newdl, int neww, int newh)
{
    undolist.Add (new UndoTotal (map, "resize", cursorx, cursory));
    teg_resizemap (&map, newdl, neww, newh);
    changed = true;
    FireEvent (EVT_MAP_TOTAL);
}

void MapWrapper::Paste (TEG_MAP *src, int l, int x, int y)
{
	undolist.Add (new UndoTotal (map, "drop floating selection", cursorx, cursory));
	teg_mapcopy (src, map, 0, 0, 0, l, x, y, src->dl, src->w, src->h);
	changed = true;
	FireEvent (EVT_MAP_CONTENT);
}

void MapWrapper::SetCursor (int x, int y)
{
	cursorx = x;
	cursory = y;
	FireEvent (EVT_CURSOR);
}

LayerWrapper::LayerWrapper() : DataWrapper()
{
    visible.clear();
    layernum = 0;
    active = 0;
}

LayerWrapper::~LayerWrapper()
{
}

bool LayerWrapper::IsVisible (int layer)
{
    return visible[layer];
}

void LayerWrapper::Reset()
{
    bool changed = false;
    std::vector<bool>::iterator i;
    
    for (i = visible.begin(); i != visible.end(); ++i)
    {
        if ((*i) == false)
        {
            (*i) = true;
            changed = true;
        }
    }
    if (changed) FireEvent(0);        
}

void LayerWrapper::SetLayerNum (int _layernum)
{
    if (layernum != _layernum)
    {        
        layernum = _layernum;
        visible.resize (layernum);
        Reset();
        FireEvent (0);
    }
}

void LayerWrapper::SetActive (int _active)
{
    if (active != _active)
    {
        active = _active;
        FireEvent (0);
    }
}

void LayerWrapper::InverseVisible (int layer)
{
    if (layer >= 0 && layer < layernum)
    {
        visible[layer] = !visible[layer];
        FireEvent(0);
    }
}

void MapBox::Copy () 
{
	int w, h, x, y, l, dl;
	if (selectState == ssSelecting)
	{
		x = selectx;
		y = selecty;
		w = selectw;
		h = selecth;
		l = selectl;
		dl = selectdl;
	}
	else // just put the single tile at cursor in the clipboard!
	{
		x = mapWrap->GetCursorX();
		y = mapWrap->GetCursorY();
		w = 1;
		h = 1;
		dl = 1;
		l = layerWrap->GetActiveLayer();
	}
	if (clipBoard != NULL)
	{
		teg_destroymap (clipBoard);
		clipBoard = NULL;
	}
	// create new map
	clipBoard = teg_createmap (dl, w, h, NULL);
	assert (clipBoard);
	// copy map section
	teg_mapcopy (mapWrap->GetMap(), clipBoard, l, x, y, 0, 0, 0, dl, w, h);
}

void MapBox::Cut ()
{
	int w, h, x, y, l, dl;
	if (selectState == ssSelecting)
	{
		x = selectx;
		y = selecty;
		w = selectw;
		h = selecth;
		l = selectl;
		dl = selectdl;
	}
	else // just put the single tile at cursor in the clipboard!
	{
		x = mapWrap->GetCursorX();
		y = mapWrap->GetCursorY();
		w = 1;
		h = 1;
		dl = 1;
		l = layerWrap->GetActiveLayer();
	}
	if (clipBoard != NULL)
	{
		teg_destroymap (clipBoard);
		clipBoard = NULL;
	}
	// create new map
	clipBoard = teg_createmap (dl, w, h, NULL);
	assert (clipBoard);
	// copy map section
	teg_mapcopy (mapWrap->GetMap(), clipBoard, l, x, y, 0, 0, 0, dl, w, h);
	mapWrap->Cut (l, x, y, dl, w, h);
}

void MapBox::Delete ()
{
	int w, h, x, y, l, dl;
	if (selectState == ssSelecting)
	{
		x = selectx;
		y = selecty;
		w = selectw;
		h = selecth;
		l = selectl;
		dl = selectdl;
	}
	else // just put the single tile at cursor in the clipboard!
	{
		x = mapWrap->GetCursorX();
		y = mapWrap->GetCursorY();
		w = 1;
		h = 1;
		dl = 1;
		l = layerWrap->GetActiveLayer();
	}
	mapWrap->Delete (l, x, y, dl, w, h);
}

void MapBox::Paste () 
{
	if (clipBoard != NULL)
	{
		if (selectState == ssSelecting)
		{
			for (int x = selectx; x < selectx + selectw; x += clipBoard->w)
			{
				for (int y = selecty; y < selecty + selecth; y += clipBoard->h)
				{
					mapWrap->Paste (clipBoard, layerWrap->GetActiveLayer(), x, y);
				}
			}
		}
		else
		{
			mapWrap->Paste (clipBoard, layerWrap->GetActiveLayer(), mapWrap->GetCursorX(), mapWrap->GetCursorY());
		}
	}
}

void MapBox::DrawClientArea (Bitmap &dest, int xofst, int yofst)
{
    GridBox::DrawClientArea (dest, xofst, yofst);
    
    TEG_MAP* map = mapWrap->GetMap();    
    if (map)
    {
        int xco, yco;
		if (selectState == ssSelecting && selectw > 0 && selecth > 0 && selectdl > 0)
		{
			// draw selection rect
			xco = selectx * GetColWidth() - xofst;
			yco = selecty * GetRowHeight() - yofst;
			int w = selectw * GetColWidth();
			int h = selecth * GetRowHeight();
			checkeredPattern.checkeredRect (dest, 
				xco - 1, yco - 1, xco + w, yco + h, cursorBlink);
		}
		// draw cursor in xor mode
        xco = mapWrap->GetCursorX() * GetColWidth() - xofst;
        yco = mapWrap->GetCursorY() * GetRowHeight() - yofst;		
        xor_mode(true);
		rect (dest, xco - 1, yco - 1, 
			xco + GetColWidth(), yco + GetRowHeight(), Color::white);
		rect (dest, xco - 2, yco - 2, 
			xco + GetColWidth() + 1, yco + GetRowHeight() + 1, Color::white);
		xor_mode (false);
    }
}
           
void MapBox::DrawCell (Bitmap &dest, int col, int row, int x, int y)
{
    TEG_MAP* map = mapWrap->GetMap();
    if (!map)
        return;

    for (int i = 0; i < map->dl; i++)
    {
        if (layerWrap->IsVisible(i))
        {
            int index;
            if (mapWrap->GetCursorX() == col && 
					mapWrap->GetCursorY() == row && 
                    cursorBlink && layerWrap->GetActiveLayer() == i)
                index = currentTile;            
            else
                index = teg_mapget (map, i, col, row);
            if (index >= 0 && index < map->tilelist->tilenum)
            {
                if (frame >= 0 && frame < map->tilelist->animsteps)
                {
                    index += frame * map->tilelist->tilenum;
                }
                teg_drawtile (dest, map->tilelist, index, x, y);
            }
        }
    }
}

MapBox::~MapBox()
{
    mapWrap->RemoveListener (&mapListener);
    layerWrap->RemoveListener (&layerListener);
    scrollWrap->RemoveListener (&scrollListener);
    if (cursorTimerId >= 0) Timer::Kill (cursorTimerId);
    if (timerId >= 0) Timer::Kill (timerId);
	if (scrollRepeatTimerId >= 0) Timer::Kill (scrollRepeatTimerId);
}

MapBox::MapBox (MapWrapper *_mapWrap, 
            LayerWrapper *_layerWrap, ScrollWrapper *_scrollWrap) : 
	GridBox(),
	layerListener(this), mapListener(this), scrollListener(this),
    blockScrollEvent (false),
    clientfocus (false)
{
	SetColNum (0);
	SetRowNum (0);
	currentTile = 0;
	currentTileR = 0;
	tool = T_BRUSH;
	cursorBlink = false;
	cursorTimerId = Timer::Install (200);
	
	mapWrap = _mapWrap;
	mapWrap->AddListener (&mapListener);
	layerWrap = _layerWrap;
	layerWrap->AddListener (&layerListener);
	scrollWrap = _scrollWrap;
	scrollWrap->AddListener (&scrollListener);
	
	frame = 0;
	timerId = -1;
	scrollRepeatTimerId = -1;
	
	selectx = 0;
	selecty = 0;
	selectw = 0;
	selecth = 0;
	selectl = 0;
	selectdl = 0;
	select_startx = 0;
	select_starty = 0;
	selectState = ssNormal;
	
	clipBoard = NULL;	
}

// called while dragging the mouse to make a selection
// or while moving the cursor by keyboard
void MapBox::UpdateSelection ()
{
	selectx = TEG_MIN (select_startx, mapWrap->GetCursorX());
	selectw = abs (mapWrap->GetCursorX() - select_startx) + 1;
	selecty = TEG_MIN (select_starty, mapWrap->GetCursorY());
	selecth = abs (mapWrap->GetCursorY()- select_starty) + 1;
}

// mouse movement, with all button states
void MapBox::MsgMousemove(const Point &d)
{
	GridBox::MsgMousemove (d);
    
	TEG_MAP* map = mapWrap->GetMap();    
	if (!map) return;
	
	Point p = GetMousePos();
	
	if (!(mouse_b & 1) && !(mouse_b & 2)) clientfocus = false;
	
	int col = GetCol (p.x() - GetClientLeft());
	int row = GetRow (p.y() - GetClientTop());
	
	if (clientfocus && col >= 0 && row >= 0 && col < map->w && row < map->h)
	{
		mapWrap->SetCursor(col, row);
	}

	if (clientfocus & IsInClientArea (p))        
	{		
		if (col >= 0 && row >= 0 && col < map->w && row < map->h)
		{
			if ((mouse_b & 1) && tool == T_BRUSH)
			{
				int currentLayer = layerWrap->GetActiveLayer();
				
				if (teg_mapget (map, currentLayer, col, row) != currentTile)
				    mapWrap->Set (currentLayer, col, row, currentTile);                
			}
			else if ((mouse_b & 2) && tool == T_BRUSH)
			{
				int currentLayer = layerWrap->GetActiveLayer();
			
				if (teg_mapget (map, currentLayer, col, row) != currentTileR)
				    mapWrap->Set (currentLayer, col, row, currentTileR);                
			}
			Redraw();
		}
	}
	// doesn't have to be in client Area
	if ((mouse_b & 1) && tool == T_SELECT && 
		selectState == ssSelecting)
	{
		if (selectState == ssSelecting)
			UpdateSelection();
		Redraw();
	}
}

// called one time per click
void MapBox::MsgRPress() 
{
    GridBox::MsgRPress();
    
    TEG_MAP* map = mapWrap->GetMap();    
    if (!map) return;

    int currentLayer = layerWrap->GetActiveLayer();
    
    Point p = GetMousePos();
    
    if (IsInClientArea (p))        
    {    
        clientfocus = true;
        int col = GetCol (p.x() - GetClientLeft());
        int row = GetRow (p.y() - GetClientTop());
    
        if (col >= 0 && row >= 0 && col < map->w && row < map->h)
        {
			mapWrap->SetCursor (col, row);
            switch (tool)
            {
                case T_SELECT:
					if (selectState == ssSelecting)
					{
						selectState = ssNormal; // cancel selection
					}
                    break;
                case T_BRUSH:
                    if (teg_mapget (map, currentLayer, col, row) != currentTileR)
                      mapWrap->Set (currentLayer, col, row, currentTileR);
                    break;
                case T_FILL:
                    if (teg_mapget (map, currentLayer, col, row) != currentTileR)
                      mapWrap->FloodFill (currentLayer, col, row, currentTileR);
                    break;
            }
            Redraw();        
        }
    }
}

void MapBox::MsgLRelease()
{
	Point p = GetMousePos();
	int col = GetCol (p.x() - GetClientLeft());
	int row = GetRow (p.y() - GetClientTop());
    TEG_MAP* map = mapWrap->GetMap();    
    if (!map) return;
	if (col >= 0 && row >= 0 && col < map->w && row < map->h)
	{
		mapWrap->SetCursor (col, row);
	}
	if (selectState == ssSelecting)
	{
		UpdateSelection (); // call updateSelection one last time
		ClearFlag (D_PRESSED);
		if (scrollRepeatTimerId >= 0) 
		{
			Timer::Kill (scrollRepeatTimerId);
			scrollRepeatTimerId = -1;
		}
	}
}

// called one time per click
void MapBox::MsgLPress()
{
    GridBox::MsgLPress();
    
    TEG_MAP* map = mapWrap->GetMap();    
    if (!map) return;

    int currentLayer = layerWrap->GetActiveLayer();
    
    Point p = GetMousePos();
	
    if (IsInClientArea (p))        
    {    
        clientfocus = true;
        int col = GetCol (p.x() - GetClientLeft());
        int row = GetRow (p.y() - GetClientTop());
    
        if (col >= 0 && row >= 0 && col < map->w && row < map->h)
        {
			mapWrap->SetCursor(col, row);
			switch (tool)
			{
				case T_SELECT:
					selectState = ssSelecting;
					selectx = col;
					selecty = row;
					selectw = 1;
					selecth = 1;
					selectdl = 1;
					selectl = currentLayer;
					select_startx = col;
					select_starty = row;
					SetFlag (D_PRESSED);
					// check each 100 msec if the mouse moves out of view
					if (scrollRepeatTimerId == -1) scrollRepeatTimerId = Timer::Install (100);
				   break;
				case T_BRUSH:
					if (teg_mapget (map, currentLayer, col, row) != currentTile)
						mapWrap->Set (currentLayer, col, row, currentTile);
					break;
				case T_FILL:
					if (teg_mapget (map, currentLayer, col, row) != currentTile)
						mapWrap->FloodFill (currentLayer, col, row, currentTile);
					break;            
			}
				
			Redraw();        
        }
    }
}

void MapBox::LayersChanged (int code)
{
    MsgDraw();
}

void MapBox::ScrollChanged (int code)
{
    if (!blockScrollEvent)
    {
        SetPos (scrollWrap->GetX() * GetColWidth(),
                scrollWrap->GetY() * GetRowHeight());
    }
}
    
void MapBox::MsgTimer (int t)
{
    TEG_MAP* map = mapWrap->GetMap();
    if (map && t >= 0 && t == timerId)
    {
        ++frame;
        if (frame >= map->tilelist->animsteps)    
            frame = 0;
        Redraw();
    }
    if (map && t == cursorTimerId)
    {
        cursorBlink = !cursorBlink;
		
        //~ int xco, yco, w, h;
		//~ if (selectState == ssSelecting && selectw > 0 && selecth > 0 && selectdl > 0)
		//~ {
			//~ // draw selection rect
			//~ xco = selectx * GetColWidth() - GetClientLeft() - 1;
			//~ yco = selecty * GetRowHeight() - GetClientTop() - 1;
			//~ w = selectw * GetColWidth() + 2;
			//~ h = selecth * GetRowHeight() + 2;
		//~ }
		//~ else
		//~ {
			//~ // draw cursor in xor mode
			//~ xco = mapWrap->GetCursorX() * GetColWidth() - GetClientLeft() - 2;
			//~ yco = mapWrap->GetCursorY() * GetRowHeight() - GetClientTop() - 2;
			//~ w = GetColWidth() + 4;
			//~ h = GetRowHeight() + 4;
		//~ }
        //~ Redraw(Rect (xco, yco, w, h));
		
		Redraw();
    }
	if (t == scrollRepeatTimerId && (mouse_b & 1) && tool == T_SELECT && 
			selectState == ssSelecting)
	{
		Point p = GetMousePos();
		if (!IsInClientArea(p))
		{
			const int jump = 32;
			if (p.x() < GetClientLeft())
			{
				SetPos (GetXOffset() - jump,  GetYOffset());
			}
			if (p.x() > GetClientLeft() + GetClientWidth())
			{
				SetPos (GetXOffset() + jump,  GetYOffset());
			}
			if (p.y() < GetClientTop())
			{
				SetPos (GetXOffset(), GetYOffset() - jump);
			}
			if (p.y() > GetClientTop() + GetClientHeight())
			{
				SetPos (GetXOffset(), GetYOffset() + jump);
			}
			int col = GetCol (p.x() - GetClientLeft());
			int row = GetRow (p.y() - GetClientTop());
		
			if (col >= 0 && row >= 0 && col < map->w && row < map->h)
			{
				mapWrap->SetCursor (col, row);
				UpdateSelection();
			}
			Redraw();
		}
	}
}

void MapBox::MapChanged (int code)
{
    TEG_MAP* map = mapWrap->GetMap();
    if (map)
    {
		if (code == MapWrapper::EVT_MAP_TOTAL)
		{
			SetColNum (map->w);
			SetRowNum (map->h);
			SetColWidth (map->tilelist->tilew);
			SetRowHeight (map->tilelist->tileh);
			blockScrollEvent = true;
			scrollWrap->SetPos (GetXOffset() / GetColWidth(), 
				GetYOffset()  / GetRowHeight());    
			scrollWrap->SetVisible (GetClientWidth() / GetColWidth(), 
				GetClientHeight() / GetRowHeight());        
			scrollWrap->SetTotal (GetColNum(), GetRowNum());
			blockScrollEvent = false;
			layerWrap->SetLayerNum (map->dl);
			if (timerId >= 0) 
			{
				Timer::Kill (timerId);
				timerId = -1;
			}
			if (map->tilelist->animsteps > 0)
			{
				timerId = Timer::Install (300);
			}
		}
		if (code == MapWrapper::EVT_CURSOR)
		{
			ScrollToCursor();
			Redraw();
		}
    }
    else
    {
        SetColNum (1);
        SetRowNum (1);
        SetColWidth (16);
        SetRowHeight (16);
        scrollWrap->SetPos (0,0);
        scrollWrap->SetVisible (0,0);
        scrollWrap->SetTotal (0,0);
        layerWrap->SetLayerNum (0);
    }    
}

void MapBox::MsgScroll()
{
    blockScrollEvent = true; // temporarily block scroll events
    scrollWrap->SetPos (GetXOffset() / GetColWidth(), GetYOffset()  / GetRowHeight());
    scrollWrap->SetVisible (GetClientWidth() / GetColWidth(), 
        GetClientHeight() / GetRowHeight());
    scrollWrap->SetTotal (GetColNum(), GetRowNum());
    blockScrollEvent = false;
}

/**
Scroll mapbox to make sure the cursor (cursorx, cursory) is visible.
*/
void MapBox::ScrollToCursor()
{
	scrollWrap->ScrollToArea (mapWrap->GetCursorX(), 
		mapWrap->GetCursorY(), 
		1, 1);
}

bool MapBox::MsgXChar(int c)
{
    TEG_MAP* map = mapWrap->GetMap();
	if (!map) return false;
	int code = c >> 8;
	int newx = mapWrap->GetCursorX(), newy = mapWrap->GetCursorY();
	switch (code)
	{
		case KEY_LEFT:
			if (key_shifts & KB_CTRL_FLAG) newx = 0; else
				if (mapWrap->GetCursorX() > 0) newx--; else newx = map->w - 1;
			break;
		case KEY_RIGHT:
			if (key_shifts & KB_CTRL_FLAG) newx = map->w - 1; else
				if (mapWrap->GetCursorX() < map->w - 1) newx++; else newx = 0;
			break;
		case KEY_UP:
			if (key_shifts & KB_CTRL_FLAG) newy = 0; else			
				if (mapWrap->GetCursorY() > 0) newy--; else newy = map->h - 1;
			break;
		case KEY_DOWN:
			if (key_shifts & KB_CTRL_FLAG) newy = map->h - 1; else			
				if (mapWrap->GetCursorY() < map->h - 1) newy++; else newy = 0;
			break;
		case KEY_HOME:
			newx = 0; newy = 0;
			break;
		case KEY_END:
			newx = map->w - 1;
			newy = map->h - 1;
			break;
	}
	
	int currentLayer = layerWrap->GetActiveLayer();
	
	switch (code)
	{
		case KEY_LEFT: case KEY_RIGHT: 
		case KEY_UP: case KEY_DOWN:
		case KEY_HOME: case KEY_END:
			if (key_shifts & KB_SHIFT_FLAG)
			{
				if (selectState == ssNormal)
				{
					selectState = ssSelecting;
					select_startx = mapWrap->GetCursorX();
					select_starty = mapWrap->GetCursorY();
					selectdl = 1;
					selectl = currentLayer;
					SetFlag (D_PRESSED); //???
				}
				if (newx > select_startx)
				{
					selectx = select_startx;
					selectw = newx - select_startx + 1;
				} else {
					selectx = newx;
					selectw = select_startx - newx + 1;
				} 
				if (newy > select_starty)
				{
					selecty = select_starty;
					selecth = newy - select_starty + 1;
				} else {
					selecty = newy;
					selecth = select_starty - newy + 1;
				}
			}
			else
			{
				selectState = ssNormal;
			}
			mapWrap->SetCursor(newx, newy);
			return true;
		case KEY_SPACE:
		{
			int cursorx = mapWrap->GetCursorX();
			int cursory = mapWrap->GetCursorY();
			switch (tool)
			{
                case T_BRUSH:
                    if (teg_mapget (map, currentLayer, cursorx, cursory) != currentTile)
                        mapWrap->Set (currentLayer, cursorx, cursory, currentTile);
                    break;
                case T_FILL:
                    if (teg_mapget (map, currentLayer, cursorx, cursory) != currentTile)
                        mapWrap->FloodFill (currentLayer, cursorx, cursory, currentTile);
                    break;            
			}
			return true;
		}
	}
	return false;
}

void MapBox::MsgGotMouse()
{
	// TODO: set cursor depending on current tool
	// setCursor ();
}

void MapBox::MsgLostMouse()
{
	// TODO: set default cursor again
}

void MapBoxMapListener::changed (int code)
{
	parent->MapChanged(code);
}

void MapBoxLayerListener::changed (int code)
{
	parent->LayersChanged(code);
}

void MapBoxScrollListener::changed (int code)
{
	parent->ScrollChanged(code);
}

