//     :     :     :     :     :     :     :     :     :     :
//   --------:  ---------:  ---------:  ---------:  ----     :
// ..|      |*..|       |*..|       |*..|       |*..|  |*....:.
//   |      |*  |       |*  |       |*  |       |*  |  |*    :
//   ---  ---*  |  |-----*  |  |-----*  |  |-----*  |  |*    :
// ....|  |*.:..|  |--****..|  |-----*..|  |--****..|  |*....:.
//     |  |* :  |    |*  :  |  |--  |*  |    |*  :  |  |*    :
//     |  |* :  |  |--*  :  |  |*|  |*  |  |--*  :  |  |*    :
// ....|  |*.:..|  |-----:..|  ---  |*..|  |-----:..|  -----.:.
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
// ....----*.:..---------*..---------*..---------*..--------*:.
//     :**** :   *********   *********   *********   ********:
//     :     :     :     :     :     :     :     :     :     :
//
// Version 0.81
// Copyright (c) 2002-2007
// Martijn "Amarillion" van Iersel.
//
// This file is part of the Tegel tilemapping add-on for allegro
//
//    Tegel is free software; you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General 
//    Public License as published by the Free Software 
//    Foundation; either version 2.1 of the License, or
//    (at your option) any later version.
//
//    Tegel is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with Tegel; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
// official website:
// http://tegel.sourceforge.net/
//
// I hope you like this program. Please send 
// questions, comments, flames and feature 
// requests to amarillion@yahoo.com.
//
#include "radarbox.h"
#include <assert.h>

using namespace MAS;

void RadarBox::GetColors()
{
    if (!map) return;
    TEG_TILELIST *tiles = map->tilelist;
    if (!tiles) return;
    
    colors.resize (tiles->tilenum);

    BITMAP *tile = create_bitmap (tiles->tilew, tiles->tileh);
    int i;
    int x, y;
    for (i = 0; i < tiles->tilenum; i++)
    {
        clear_bitmap (tile);
        teg_drawtile (tile, tiles, i, 0, 0);
        int r, g, b;
        r = 0; g = 0; b = 0;
        for (y = 0; y < tile->h; ++y)
        {
            for (x = 0; x < tile->w; ++x)
            {
                int color = getpixel (tile, x, y);
                r += getr (color);
                g += getg (color);
                b += getb (color);
            }
        }
        r /= (tile->w * tile->h);            
        g /= (tile->w * tile->h);            
        b /= (tile->w * tile->h);            
        colors[i] = makecol (r, g, b);            
    }
    
    destroy_bitmap (tile);    
}

RadarBox::RadarBox(MapWrapper *_mapWrap, ScrollWrapper *_scrollWrap) : 
    MyScrollBox (), mapListener(this), scrollListener(this), magnification (1)
{
    mapWrap = _mapWrap;
    mapWrap->AddListener (&mapListener);    
    scrollWrap = _scrollWrap;
    scrollWrap->AddListener (&scrollListener);
    map = NULL;    
    radar = NULL;
    cursorx = 0;
    cursory = 0;
    cursorw = 0;
    cursorh = 0;
}

// called when either map has changed, or radarbox is resized
void RadarBox::UpdateMagnification()
{
    magnification = 1;
    if (!radar) return;
    if (radar->w == 0 || radar->h == 0) return;
    
    while (radar->w * (magnification+1) < GetClientWidth() &&
        radar->h * (magnification+1) < GetClientHeight())
            magnification++;    
}

void RadarBox::UpdateSize()
{
    MyScrollBox::UpdateSize();
    UpdateMagnification();
}

RadarBox::~RadarBox()
{
    mapWrap->RemoveListener (&mapListener);
    scrollWrap->RemoveListener (&scrollListener);
	destroy_bitmap (radar);
}

void RadarBox::DrawClientArea (Bitmap &dest, int xofst, int yofst)
{
    if (!radar) return;
    
    stretch_blit (radar, dest, 0, 0, radar->w, radar->h, 
        -xofst, -yofst, radar->w * magnification, radar->h * magnification);
    //blit (radar, dest, xofst, yofst, 
    //            0, 0, dest.w(), dest.h());
                
    if (cursorw > 0 && cursorh > 0)
    {
		xor_mode (true);
        rect (dest, 
            (cursorx * magnification) - xofst, 
            (cursory * magnification) - yofst, 
            - xofst + (cursorx + cursorw) * magnification, 
            - yofst + (cursory + cursorh) * magnification, 
            Color::white);
		xor_mode (false);
    }    
}

void RadarBox::MsgClick()
{
    if (!map) return;
    
    Point p = GetMousePos();
    
    if (IsInClientArea (p))
    {
    
        int mx = (p.x() - GetClientLeft() + GetXOffset()) / magnification;
        int my = (p.y() - GetClientTop() + GetYOffset()) / magnification;
    
        if (mx >= 0 && my >= 0 && mx < map->w && my < map->h)
        {
            // try to center cursor around x, y
    
            int newcursorx = mx - (cursorw / 2);
            int newcursory = my - (cursorh / 2);
        
            if (newcursorx < 0) newcursorx = 0;
            if (newcursory < 0) newcursory = 0;
            if (newcursorx + cursorw >= map->w) newcursorx = map->w - cursorw;
            if (newcursory + cursorh >= map->h) newcursory = map->h - cursorh;

            scrollWrap->SetPos (newcursorx, newcursory);
        }
    }
}

void RadarBox::MapChanged (int code)
{
    // check if map pointer has changed
    if (mapWrap->GetMap() != map)
    {
        map = mapWrap->GetMap();
        if (radar) 
        {
            destroy_bitmap (radar);
            radar = NULL;
        }
        if (map)
        {
            radar = create_bitmap (map->w, map->h);
            GetColors();
           
            int x, y;
            for (y = 0; y < map->h; ++y)
            {   
                for (x = 0; x < map->w; ++x)
                {
                    int index = teg_mapget (map, 0, x, y);
                    putpixel (radar, x, y, colors[index]);
                }
            }

            UpdateMagnification();
            SetScrollSize (radar->w * magnification, radar->h * magnification);
            
            Redraw();
        }
    }
    // if not, then map data has changed
    else if (map)
    {
        int x, y;
        for (y = 0; y < map->h; ++y)
        {   
            for (x = 0; x < map->w; ++x)
            {
                int index = teg_mapget (map, 0, x, y);
                putpixel (radar, x, y, colors[index]);
            }
        }
        Redraw();
    }
}

void RadarBox::ScrollChanged (int code)
{
    cursorx = scrollWrap->GetX();     
    cursory = scrollWrap->GetY();     
    cursorw = scrollWrap->GetXVis();     
    cursorh = scrollWrap->GetYVis();     
    Redraw();
}

void RadarBoxMapListener::changed (int code)
{
	parent->MapChanged(code);
}

void RadarBoxScrollListener::changed (int code)
{
	parent->ScrollChanged(code);
}
