//     :     :     :     :     :     :     :     :     :     :
//   --------:  ---------:  ---------:  ---------:  ----     :
// ..|      |*..|       |*..|       |*..|       |*..|  |*....:.
//   |      |*  |       |*  |       |*  |       |*  |  |*    :
//   ---  ---*  |  |-----*  |  |-----*  |  |-----*  |  |*    :
// ....|  |*.:..|  |--****..|  |-----*..|  |--****..|  |*....:.
//     |  |* :  |    |*  :  |  |--  |*  |    |*  :  |  |*    :
//     |  |* :  |  |--*  :  |  |*|  |*  |  |--*  :  |  |*    :
// ....|  |*.:..|  |-----:..|  ---  |*..|  |-----:..|  -----.:.
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
//     |  |* :  |       |*  |       |*  |       |*  |      |*:
// ....----*.:..---------*..---------*..---------*..--------*:.
//     :**** :   *********   *********   *********   ********:
//     :     :     :     :     :     :     :     :     :     :
//
// Version 0.81
// Copyright (c) 2002-2007
// Martijn "Amarillion" van Iersel.
//
// This file is part of the Tegel tilemapping add-on for allegro
//
//    Tegel is free software; you can redistribute it and/or modify
//    it under the terms of the GNU Lesser General 
//    Public License as published by the Free Software 
//    Foundation; either version 2.1 of the License, or
//    (at your option) any later version.
//
//    Tegel is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public License
//    along with Tegel; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
// official website:
// http://tegel.sourceforge.net/
//
// I hope you like this program. Please send 
// questions, comments, flames and feature 
// requests to amarillion@yahoo.com.
//
#include <assert.h>
#include <stdio.h>
#include "tegel.h"
#include <memory.h>
#include "tegfile.h"
#include <ctype.h>

#define TEGERR_OK 0
#define TEGERR_TLL_VERSION 1
#define TEGERR_MAP_VERSION 2
#define TEGERR_RESOURCE_NOT_FOUND 3
#define TEGERR_RESOURCE_FORMAT 4
#define TEGERR_HEADER_FORMAT 5
#define TEGERR_UEO_BODY 6
#define TEGERR_MEMORY 7
#define TEGERR_UEO_HEADER 8
#define TEGERR_BPP_IMPOS 9
#define TEGERR_FOPEN 10
#define TEGERR_DATAFILE 11
#define TEGERR_DATAOBJ 12

char *teg_errorlist[] = 
    {
	"Everything is ok", 
	"The tilelist version is not supported",
	"The map version is not supported",
	"Could not find required resource [%s] in datafile",
	"Required resource [%s] is of the wrong size or format",
	"The tilelist header does not contain required fields",
	"Unexpected end of tilelist body",
	"Out of memory!",
        "Unexpected end of tilelist header",
        "Impossible number of bits per pixel",
        "file can't be opened",
		"General datafile error: %s",
		"Error with tegel datafile object: %s"
    };


char teg_error[256] = "Everything is ok";

int _teg_installed = 0;

/*
  load map from file

  map file format:
  [4]header = MIMP
  [4]version = 1
  [4]w
  [4]h
  [4]dl
  [4 * dl * w * h] data
*/
TEG_MAP *teg_loadmap_from_stream (file_t f, TEG_TILELIST *tilelist)
{
    char buf[4];
    int version = 0;
    int error = FALSE;
    int totalsize = 0;
    TEG_MAP *result = malloc (sizeof(TEG_MAP));
    if (result != NULL)
    {
        result->tilelist = tilelist;
        if (!error) error = (file_read (f, buf, 4) != 4);
        if (memcmp (buf, "MIMP", 4) != 0) error = TRUE;
        if (!error) error = (file_read (f, &version, 4) != 4);
        if (version != 1) 
        {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[2]);         
        }        
        if (!error) error = (file_read (f, &result->w, 4) != 4);
        if (!error) error = (file_read (f, &result->h, 4) != 4);
        if (!error) error = (file_read (f, &result->dl, 4) != 4);
        totalsize = sizeof(int) * result->w * result->h * result->dl;
        result->data = NULL;
        if (!error) result->data = malloc (totalsize);
        if (result->data == NULL) error = TRUE;
        if (!error) error = (file_read (f, result->data, totalsize) != totalsize);
        if (error)
        {
           if (result->data != NULL) free (result->data);
           free (result);
           result = NULL;
        }        
    }
    return result;
}


TEG_MAP *teg_loadmap (const char *filename, TEG_TILELIST *tilelist)
{
    file_t inf;
    TEG_MAP *result = NULL;
    int error = FALSE;

    file_set_driver (stdiofile_driver);
    inf = file_open(filename, FFLAG_READ);
    if (inf != NULL)
    {
        result = teg_loadmap_from_stream (inf, tilelist);
        if (!result)
        {
            error = TRUE;
        }
    }
    else 
    {
        error = TRUE;
        strcpy (teg_error, teg_errorlist[TEGERR_FOPEN]);
    }
    file_close (inf);
    return result;
}


void *load_teg_tilelist (PACKFILE *f, long size)
{
    int error = FALSE;
    TEG_TILELIST *result = malloc (sizeof (TEG_TILELIST));
    void *p = malloc (size);
    if (result == NULL || p == NULL)
    {
        error = TRUE;
    }    
    if (!error)
    {
        pack_fread (p, size, f);
        if (pack_ferror(f))
        {
            error = TRUE;
        }
    }
    if (!error)
    {
        result->tilew = 0;
        result->tilenum = 0;
        result->tileh = 0;
        result->bpp = 0;
        result->tiles = NULL;
        result->largebmp = NULL;
        result->animsteps = 0;
        result->flags = 0;
        result->rawdata = p;
        result->rawsize = size;
    }        
    if (error)
    {
        free (p);
        free (result);
        result = NULL;
        p = NULL;
    }
    return result;
}

//~ // read a line, Discard everything after a # character
//~ // if the line is empty, skip it and go to the next line.
//~ int teg_readline_without_comments (file_t f, void * buffer, int size)
//~ {
	
//~ }

// grow or shrink the tilelist, preserving as much data as possible
// returns 0 when failed, 1 on success.
int teg_tilelist_resize (TEG_TILELIST *tiles, int newtilenum, int newanimsteps)
{
	// try to allocate new memory
	TEG_TILE *temp;
	assert (tiles);
	assert (newtilenum >= 0);
	assert (newanimsteps >= 0);
	temp = malloc (newtilenum * newanimsteps * sizeof (TEG_TILE));	
	if (temp)
	{
		int frame;
		// set mem to 0
		memset (temp, 0, newtilenum * newanimsteps * sizeof (TEG_TILE));
		// copy old values
		for (frame = 0; frame < TEG_MIN (newanimsteps, tiles->animsteps); ++frame)
		{
			memcpy ((temp + frame * newtilenum), (tiles->tiles + frame * tiles->tilenum),
			    TEG_MIN (newtilenum, tiles->tilenum) * sizeof (TEG_TILE));
		}
		// free old memory
		free (tiles->tiles);
		tiles->tiles = NULL;
		// replace old with new
		tiles->tiles = temp;
		tiles->animsteps = newanimsteps;
		tiles->tilenum = newtilenum;
		return 1;
	}
	else
	{
		strcpy (teg_error, teg_errorlist[TEGERR_MEMORY]);
		return 0;		
	}
}

// load data
// first two lines have already been processed
TEG_TILELIST *teg_loadtiles_data_v1 (file_t f, const DATAFILE *data)
{
    TEG_TILELIST *temp;
    int error = FALSE;
    char buffer[256];

    int i;
    int type;
    DATAFILE* obj;
    
    //*******************************************
    // first we read the header of the tilelist
    //*******************************************
    
    temp = malloc (sizeof (TEG_TILELIST));
    if (!temp) {
        error = TRUE;
        strcpy (teg_error, teg_errorlist[TEGERR_MEMORY]);
    }
    if (!error)
    {
        temp->rawdata = NULL;
        temp->rawsize = 0;
    }

    
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " tilew = %d ", &temp->tilew) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " tileh = %d ", &temp->tileh) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " bpp = %d ", &temp->bpp) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
        if (!(temp->bpp == 8 || temp->bpp == 15 || 
            temp->bpp == 16 || temp->bpp == 24 || temp->bpp != 32)) 
        {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_BPP_IMPOS]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " flags = %d ", &temp->flags) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " animsteps = %d ", &temp->animsteps) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " tilenum = %d ", &temp->tilenum) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    
    //*******************************************
    //now the body
    //*******************************************
    
    type = (temp->flags & 3);

    // read largebmp if applicable
    if (!error && (type == TEG_TILEFLAG_LARGE || type == TEG_TILEFLAG_LARGEI))
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, "%79[^\n]\n", buffer) != 1)
        {                  
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);
        }
        if (!error) 
        {
            obj = find_datafile_object (data, buffer);                    
            if (!obj)
            {
                error = TRUE;
                sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
            }    
        }
        if (obj != NULL)
        {
            if (obj->type == DAT_BITMAP)
            {               
                temp->largebmp =  (BITMAP*)obj->dat;
            }
            else 
            {
                error = TRUE;
                sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
            }            
        }
        else 
        {
            error = TRUE;
            sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
        }
    }

    // allocate memory for the tile pointers
    temp->tiles = NULL;
    if (!error && (type == TEG_TILEFLAG_LARGEI || type == TEG_TILEFLAG_BMP ||
            type == TEG_TILEFLAG_RLE))
    {
        temp->tiles = malloc (temp->tilenum * temp->animsteps * sizeof(TEG_TILE));
        if (temp->tiles == NULL)
        {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_MEMORY]);
        }
    }    
    
    // read the tile pointers if applicable
    if (!error && 
        (type == TEG_TILEFLAG_BMP || type == TEG_TILEFLAG_RLE))
    {
            for (i = 0; 
                 (i < (temp->tilenum * temp->animsteps)) && (error == FALSE); 
                 i++)
            {                
                file_gets_ex (f, buffer, sizeof(buffer));        
                if (sscanf (buffer, "%79[^\n]\n", buffer) != 1)
                {                  
                    error = TRUE;
                    strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);
                }
                if (!error) 
                {
                    obj = find_datafile_object (data, buffer);                    
                    if (!obj)
                    {
                       error = TRUE;
                       sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
                    }    
                }
                if (!error)
                {
                    if (type == TEG_TILEFLAG_BMP &&
                        obj->type == DAT_BITMAP)
                    {
                        BITMAP *bmp = (BITMAP*)obj->dat;
                        if (bmp->w == temp->tilew
                            && bmp->h == temp->tileh)
                        {
                            temp->tiles[i].bmp = bmp;
                            temp->tiles[i].rle = NULL;
                            temp->tiles[i].index = 0;
                            temp->tiles[i].flags = 0;
						}
                        else 
                        {
                           error = TRUE;
                           sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
                        }                        
                    }
                    if (type == TEG_TILEFLAG_RLE &&
                        obj->type == DAT_RLE_SPRITE)
                    {
                        RLE_SPRITE *rle = (RLE_SPRITE*)obj->dat;
                        if (rle->w == temp->tilew
                            && rle->h == temp->tileh)
                        {
                            temp->tiles[i].rle = rle;
                            temp->tiles[i].bmp = NULL;
                            temp->tiles[i].index = 0;
							temp->tiles[i].flags = 0;
                        }
                        else 
                        {
                           error = TRUE;
                           sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
                        }                        
                    }
                }               
            }
    }
    if (!error && (type == TEG_TILEFLAG_LARGEI))
    {
                // now read the indices
                for (i = 0; 
                    (i < (temp->tilenum * temp->animsteps)) && (error == FALSE); 
                         i++)
                {
                    // read a line
                    int index;
                    
                    file_gets_ex (f, buffer, sizeof(buffer));        
                    if (sscanf (buffer, " %d ", &index) == 1)
                    {
                        temp->tiles[i].rle = NULL;
                        temp->tiles[i].bmp = NULL;
                        temp->tiles[i].index = index;
						temp->tiles[i].flags = 0;
                    }
                    else 
                    {
                        error = TRUE;                    
                        strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);
                    }
                                        
                }                
    }        
    // clean memory up if there was an error
    if (error)
    {
        if (temp)
        {
            free (temp->tiles);
            temp->tiles = NULL;
            free (temp);
            temp = NULL;
        }
    }
    return temp;
}

// read TEG_TILE
// returns 0 on failure -> error is set in teg_error
// returns 1 on success
// data returned in tempframes buffer (should be allocated and freed by caller)
// tiles is used to read animsteps and flags from. It is not changed in this function
// datafile is used to search for the bitmap / rle sprite
int teg_scan_data_line (char *line, TEG_TILE *tempframes, TEG_TILELIST *tiles, const DATAFILE *data)
{
	int pos = 0;
	int scanned;
	int lastframe = FALSE;
	int frame;
	int type = (tiles->flags & 3);
	char buffer[256];
	int error = FALSE;
	int flags = 0;	
	if (sscanf (line + pos, " %i%*[ \t.,:;]%n", &flags, &scanned) > 0)
	{
		pos += scanned;
	}	
	else
	{
		strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);
		error = TRUE;
	}
	for (frame = 0; frame < tiles->animsteps && !lastframe && !error; )
	{	
		tempframes[frame].rle = NULL;
		tempframes[frame].bmp = NULL;
		tempframes[frame].index = 0;
		tempframes[frame].flags = flags;	
		if (type == TEG_TILEFLAG_LARGEI)
		{
			int index;
			if (sscanf (line + pos, " %d%n", &index, &scanned) > 0)
			{				
				pos += scanned;		
				tempframes[frame].index = index;				
			}
			else 
			{				
				lastframe = TRUE;				
			}								
		}        
		else
		{
			// now read a tile-id. if none is found, assume we've read the last tile-id of this line
			if (sscanf (line + pos, " %255[0-9a-zA-Z_]%n", buffer, &scanned) < 1)
			{
				lastframe = TRUE;				
			}
			else
			{
				DATAFILE *obj = NULL;
				pos += scanned;		
				// search for a tile by this id
				if (!error) 
				{
					obj = find_datafile_object (data, buffer);                    
					if (!obj)
					{
					   error = TRUE;
					   sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
					}    
				}
				if (!error)
				{
					if (type == TEG_TILEFLAG_BMP && obj->type == DAT_BITMAP)
					{
						BITMAP *bmp = (BITMAP*)obj->dat;
						if (bmp->w == tiles->tilew
							&& bmp->h == tiles->tileh)
						{
							tempframes[frame].bmp = bmp;							
						}
						else 
						{
						   error = TRUE;
						   sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
						}                        
					}
					if (type == TEG_TILEFLAG_RLE && obj->type == DAT_RLE_SPRITE)
					{
						RLE_SPRITE *rle = (RLE_SPRITE*)obj->dat;
						if (rle->w == tiles->tilew
							&& rle->h == tiles->tileh)
						{
							tempframes[frame].rle = rle;							
						}
						else 
						{
						   error = TRUE;
						   sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
						}                        
					}
				}               
			}
		}
		if (!error && !lastframe) frame++;
		// if we're after the first, search for a separator
		// if none is found, assume we've read the last tile-id of this line.
		if (!error) 
		{
			scanned = 0;
			sscanf (line + pos, "%*[ \t.,:;]%n", &scanned);
			if (scanned <= 0)
			{
				// no separator found
				lastframe = TRUE;
			}
			else
			{
				pos += scanned;
			}
		}
	}
	// if there were no frames, there is a problem
	if (!error && frame == 0)
	{
		error = TRUE;
		strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);	
	} 
	if (!error && frame < tiles->animsteps)
	{
		// fill out the rest
		// by looping throught the ones already set
		// so if the line reads "a, b, c" and animsteps is 8, the result will be
		// a, b, c, a, b, c, a, b
		int i;
		for (i = frame; i < tiles->animsteps; ++i)
		{
			memcpy (tempframes + i, tempframes + (i % frame), sizeof (TEG_TILE));
		}			
	}
	return (error == FALSE);
}

// load data
// first two lines have already been processed
// version 2
/* 
differences with version 1:
tilenum = xxx line in header obsolete, inferred from length of list
all data for one tile is on a single line, separated by tabs, spaces or commas. 
the file can have comments starting with #
tiles can have flags or properties now
*/
TEG_TILELIST *teg_loadtiles_data_v2 (file_t f, const DATAFILE *data)
{
    TEG_TILELIST *temp;
    int error = FALSE;
    char buffer[256];

    int i;
    int type;
	int tilecount;
    DATAFILE* obj;
    
    //*******************************************
    // first we read the header of the tilelist
    //*******************************************
    
    temp = malloc (sizeof (TEG_TILELIST));
    if (!temp) {
        error = TRUE;
        strcpy (teg_error, teg_errorlist[TEGERR_MEMORY]);
    }
    if (!error)
    {
        temp->rawdata = NULL;
        temp->rawsize = 0;
    }    
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " tilew = %d ", &temp->tilew) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " tileh = %d ", &temp->tileh) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " bpp = %d ", &temp->bpp) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
        if (!(temp->bpp == 8 || temp->bpp == 15 || 
            temp->bpp == 16 || temp->bpp == 24 || temp->bpp != 32)) 
        {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_BPP_IMPOS]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " flags = %d ", &temp->flags) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, " animsteps = %d ", &temp->animsteps) != 1) {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
	// temp->tilenum is no longer read from file but inferred from list length
	temp->tilenum = 0;
	temp->tiles = NULL;
    tilecount = 0; // number of tiles we have encountered thus far.
    
    //*******************************************
    //now the body
    //*******************************************
    
    type = (temp->flags & 3);
    // read largebmp if applicable
    if (!error && (type == TEG_TILEFLAG_LARGE || type == TEG_TILEFLAG_LARGEI))
    {
        file_gets_ex (f, buffer, sizeof(buffer));        
        if (sscanf (buffer, "%79[^\n]\n", buffer) != 1)
        {                  
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_UEO_BODY]);
        }
        if (!error) 
        {
            obj = find_datafile_object (data, buffer);                    
            if (!obj)
            {
                error = TRUE;
                sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
            }    
        }
        if (obj != NULL)
        {
            if (obj->type == DAT_BITMAP)
            {               
                temp->largebmp =  (BITMAP*)obj->dat;
				// determine # of tiles in case of TEG_TILEFLAG_LARGE
				if (type == TEG_TILEFLAG_LARGE && temp->tilew > 0 && temp->tileh > 0)
				{
					temp->tilenum = (temp->largebmp->w / temp->tilew) * (temp->largebmp->h / temp->tileh);
				}
            }
            else 
            {
                error = TRUE;
                sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_FORMAT], buffer);
            }            
        }
        else 
        {
            error = TRUE;
            sprintf (teg_error, teg_errorlist[TEGERR_RESOURCE_NOT_FOUND], buffer);
        }
    }
    
    // read the tile pointers if applicable
    if (!error && 
        (type == TEG_TILEFLAG_BMP || type == TEG_TILEFLAG_RLE || type == TEG_TILEFLAG_LARGEI))
    {
		TEG_TILE* tempframes = malloc (temp->animsteps * sizeof (TEG_TILE));
		int last = FALSE;
		assert (tempframes);		
		for (i = 0; !error && !last; i++)
		{
			file_gets_ex (f, buffer, sizeof(buffer));
			if (buffer[0] == '!') // a "!" at the beginning of a line marks the last tile
			{
				last = TRUE;
			}
			else
			{
				if (!teg_scan_data_line (buffer, tempframes, temp, data))
				{
					error = TRUE;
					// error message already set by teg_scan_data_line
				}
				else
				{
					// add the tempframes
					// resize if necessary
					int j;
					if (tilecount >= temp->tilenum)
					{
						// over-allocate a bit so we won't need to do this every loop
						teg_tilelist_resize (temp, temp->tilenum + 20, temp->animsteps);
					}
					// copy the data
					for (j = 0; j < temp->animsteps; ++j)
					{
						memcpy (temp->tiles + i + j * temp->tilenum, tempframes + j, sizeof (TEG_TILE));
					}
					tilecount++;
				}			
			}
		}			
		// now finally resize back to the number of tiles actually in use.
		if (!error) teg_tilelist_resize (temp, tilecount, temp->animsteps);
		// clean up temporary allocation
		free (tempframes);
		tempframes = NULL;
    }    
    // clean memory up if there was an error
    if (error)
    {
        if (temp)
        {
            free (temp->tiles);
            temp->tiles = NULL;
            free (temp);
            temp = NULL;
        }
    }
    return temp;
}

TEG_TILELIST *teg_loadtiles_data_v3 (file_t f, const DATAFILE *data)
{
	strcpy (teg_error, teg_errorlist[TEGERR_TLL_VERSION]);
	return NULL;
}

TEG_TILELIST *teg_loadtiles_from_stream (file_t f, const DATAFILE *data)
{
	int error = FALSE;
    char buffer[256];

	int version = 0;
	
    // read TILELIST
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));
        if (strcmp (buffer, "TILELIST\n") != 0)
        {
            error = TRUE; // not right format: doesn't start with TILELIST
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
    if (!error)
    {
        file_gets_ex (f, buffer, sizeof(buffer));
        if (sscanf (buffer, " version = %d ", &version) != 1)
        {
            error = TRUE;
            strcpy (teg_error, teg_errorlist[TEGERR_HEADER_FORMAT]);
        }
    }
	if (!error)
	{        
		if (version == 1)
		{
			return teg_loadtiles_data_v1 (f, data);
		} 
		else if (version == 2)
		{
			return teg_loadtiles_data_v2 (f, data);
		}			
		else 
		{
            error = TRUE; // not the right format version
            strcpy (teg_error, teg_errorlist[TEGERR_TLL_VERSION]);
        }
	}
	return NULL;	
}

void *load_teg_map (PACKFILE *f, long size)
{
    int error = FALSE;    
    void *p = malloc (size);
    TEG_MAP *result = NULL;
    file_t fmem;
    if (p == NULL)
    {
        error = TRUE;
    }    
    if (!error)
    {
        pack_fread (p, size, f);
        if (pack_ferror(f))
        {
            error = TRUE;
        }
    }    
    if (!error)
    {
        fmem = memfile_openbuf (p, size);
        result = teg_loadmap_from_stream (fmem, NULL);
        file_close (fmem);
        if (!result) error = TRUE;
    }
    if (error)
    {
        free (p);
        free (result);
        result = NULL;
        p = NULL;
    }
    return result;
}

void destroy_teg_tilelist (void *data)
{
    TEG_TILELIST *tiles = (TEG_TILELIST*)(data);
    teg_destroytiles (tiles);
}

void destroy_teg_map (void *data)
{
    teg_destroymap ((TEG_MAP*)data);
}

DATAFILE *teg_fixup_datafile (DATAFILE *data)
{
    // search for tll objs
	DATAFILE *iter = data;
    while (iter->type != DAT_END)
    {
        // open a memory stream and load them again.
        if (iter->type == DAT_TEG_TILELIST)
        {
            file_t fmem;
            TEG_TILELIST* tiles = (TEG_TILELIST*)(iter->dat);
            TEG_TILELIST *temp;
            fmem = memfile_openbuf (tiles->rawdata, tiles->rawsize);
            temp = (void*)teg_loadtiles_from_stream (fmem, data);
            file_close (fmem);
            if (temp)
            {
                teg_destroytiles (tiles);
                iter->dat = temp;
            }
			else
			{
				// loading of datafile object failed, entire datafile should fail now.
				unload_datafile (data);
				return NULL;
			}
        }
        ++iter;
    }    
	return data;
}

DATAFILE *teg_load_datafile (const char *datafile)
{
    DATAFILE *result;    
    if (!_teg_installed) return NULL; // error, install first!
    result = load_datafile (datafile);
    if (!result) 
	{
		sprintf (teg_error, teg_errorlist[TEGERR_DATAFILE], allegro_error);
	}
	else
	{
		result = teg_fixup_datafile (result); // may fail...
		if (!result)
		{
			char buffer[sizeof(teg_error)];
			strcpy (buffer, teg_error);
			sprintf (teg_error, teg_errorlist[TEGERR_DATAOBJ], buffer);
		}
	}	
    return result;
}

void teg_unload_datafile (DATAFILE *data)
{
    unload_datafile (data);
}

int teg_install()
{
    if (_teg_installed == 0)
    {
        register_datafile_object (DAT_TEG_TILELIST, load_teg_tilelist, destroy_teg_tilelist);
        register_datafile_object (DAT_TEG_MAP, load_teg_map, destroy_teg_map);
        _teg_installed = 1;
    }
    return 0;
}

void teg_uninstall()
{
}

// create empty map
// set everything to 0
TEG_MAP *teg_createmap (int dl, int w, int h, TEG_TILELIST *tilelist)
{
    TEG_MAP *result = malloc(sizeof(TEG_MAP));
    if (result != NULL)
    {
        result->data = malloc(dl * w * h * sizeof(int));
        if (result->data != NULL)
        {
            result->dl = dl;
            result->w = w;
            result->h = h;
            result->tilelist = tilelist;
            memset (result->data, 0, dl * w * h * sizeof(int));
        }
        else
        {
            // if failed to allocate room for data,
            // also destroy the allocated struct.
            free (result);
            result = NULL;
        }
    }
    return result;
}


// destroy a map
void teg_destroymap (TEG_MAP* map)
{
    if (map != NULL)
    {
        free (map->data);
    }
    free (map);
}

// get index at a certain pos in the map
int teg_mapget (const TEG_MAP* map, int l, int x, int y)
{
    assert(map != NULL);
    assert(l >= 0 && l < map->dl);
    assert(x >= 0 && x < map->w);
    assert(y >= 0 && y < map->h);
    assert(map->data != NULL);
    return (map->data[x + map->w * (y + map->h * l)]);
}
// put index at a certain pos in the map
void teg_mapput (TEG_MAP* map, int l, int x, int y, int val)
{
    assert(map != NULL);
    assert(l >= 0 && l < map->dl);
    assert(x >= 0 && x < map->w);
    assert(y >= 0 && y < map->h);
    assert(map->data != NULL);
    map->data[x + map->w * (y + map->h * l)] = val;
}

// loads in tiles and links with bitmaps found in a datafile
// returns NULL if an error occured, either in the loading
// or in the pointer matching process.
TEG_TILELIST *teg_loadtiles (const char *filename, const DATAFILE *data)
{
    file_t inf;
    TEG_TILELIST *result = NULL;
    int error = FALSE;
    
    file_set_driver (stdiofile_driver);
    inf = file_open(filename, FFLAG_READ);
    if (inf != NULL)
    {
        result = teg_loadtiles_from_stream (inf, data);           
        if (!result)
        {
            error = TRUE;
        }
    }
    else 
    {
        error = TRUE;
        strcpy (teg_error, teg_errorlist[TEGERR_FOPEN]);
    }
    file_close (inf);
    return result;
}

// destroy a tilelist.
void teg_destroytiles (TEG_TILELIST* tiles)
{    
    if (tiles != NULL)
    {
        free (tiles->tiles);
        free (tiles->rawdata);
    }
    free (tiles);
}

// draws the tilemap scrolled by xview and yview.
void teg_draw (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview)
{
    assert (map != NULL);
    assert (bmp != NULL);
    assert (map->tilelist != NULL);

    switch (map->tilelist->flags & 3)
    {
    case TEG_TILEFLAG_BMP: teg_draw_bmp (bmp, map, layer, xview, yview, 0); break;
    case TEG_TILEFLAG_RLE: teg_draw_rle (bmp, map, layer, xview, yview, 0); break;
    case TEG_TILEFLAG_LARGE: teg_draw_large (bmp, map, layer, xview, yview, 0); break;
    case TEG_TILEFLAG_LARGEI: teg_draw_largei (bmp, map, layer, xview, yview, 0); break;
    }
}

void teg_draw_frame (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview, int frame)
{
    assert (map != NULL);
    assert (bmp != NULL);
    assert (map->tilelist != NULL);
    assert (frame < map->tilelist->animsteps);

    switch (map->tilelist->flags & 3)
    {
    case TEG_TILEFLAG_BMP: teg_draw_bmp (bmp, map, layer, xview, yview, frame); break;
    case TEG_TILEFLAG_RLE: teg_draw_rle (bmp, map, layer, xview, yview, frame); break;
    case TEG_TILEFLAG_LARGE: teg_draw_large (bmp, map, layer, xview, yview, frame); break;
    case TEG_TILEFLAG_LARGEI: teg_draw_largei (bmp, map, layer, xview, yview, frame); break;
    }
}

// specialised funcs for rle, bmp, large or largei tilelists
void teg_draw_rle (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview, int frame)
{
    int x, y;
    int tilex, tiley;

    for (tilex = 0; tilex < map->w; tilex++)
    {
        for (tiley = 0; tiley < map->h; tiley++)
        {
            RLE_SPRITE *s;
            int i;
            x = tilex * map->tilelist->tilew - xview;
            y = tiley * map->tilelist->tileh - yview;
            i = teg_mapget(map, layer, tilex, tiley);
            if (i >= 0 && i < map->tilelist->tilenum)
            {
            	i += frame * map->tilelist->tilenum;
                s = map->tilelist->tiles[i].rle;
                if (s != NULL)
                    draw_rle_sprite (bmp, s, x, y);
            }
        }
    }
}

void teg_draw_bmp (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview, int frame)
{
    int x, y;
    int tilex, tiley;

    for (tilex = 0; tilex < map->w; tilex++)
    {
        for (tiley = 0; tiley < map->h; tiley++)
        {
            x = tilex * map->tilelist->tilew - xview;
            y = tiley * map->tilelist->tileh - yview;
            draw_sprite (bmp,
                map->tilelist->tiles[teg_mapget(map, layer, tilex, tiley)].bmp,
                x, y);
        }
    }
}

void teg_draw_large (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview, int frame)
{
    int x, y;
    int tilex, tiley;
    int tiles_per_row = map->tilelist->largebmp->w / map->tilelist->tilew;
    int index;

    for (tilex = 0; tilex < map->w; tilex++)
    {
        for (tiley = 0; tiley < map->h; tiley++)
        {
            x = tilex * map->tilelist->tilew - xview;
            y = tiley * map->tilelist->tileh - yview;
            index = teg_mapget(map, layer, tilex, tiley);
            masked_blit (map->tilelist->largebmp,
                bmp,
                (index % tiles_per_row) * map->tilelist->tilew,
                (index / tiles_per_row) * map->tilelist->tileh,
                x, y,
                map->tilelist->tilew, map->tilelist->tileh);
        }
    }
}

void teg_draw_largei (BITMAP *bmp, const TEG_MAP* map, int layer, int xview, int yview, int frame)
{
    int x, y;
    int tilex, tiley;
    int tiles_per_row = map->tilelist->largebmp->w / map->tilelist->tilew;
    int temp, index;

    for (tilex = 0; tilex < map->w; tilex++)
    {
        for (tiley = 0; tiley < map->h; tiley++)
        {
            x = tilex * map->tilelist->tilew - xview;
            y = tiley * map->tilelist->tileh - yview;
            temp = teg_mapget(map, layer, tilex, tiley);
            if (temp >= 0 && temp < map->tilelist->tilenum)
            {
            	temp += frame * map->tilelist->tilenum;
                index = map->tilelist->tiles[temp].index;
                masked_blit (map->tilelist->largebmp,
                    bmp,
                    (index % tiles_per_row) * map->tilelist->tilew,
                    (index / tiles_per_row) * map->tilelist->tileh,
                    x, y,
                    map->tilelist->tilew, map->tilelist->tileh);
            }
        }
    }
}

// idem as teg_draw, but only a part of the target bitmap will be drawn.
// x, y, w and h are relative to the target bitmap coordinates
// xview and yview are relative to the target bitmap (0,0), not to (x,y)
void teg_partdraw (BITMAP *bmp, const TEG_MAP* map, int layer, int x, int y, int w, int h, int xview, int yview)
{
}

void teg_partdraw_rle (BITMAP *bmp, const TEG_MAP* map, int layer, int x, int y, int w, int h, int xview, int yview)
{
}

void teg_partdraw_bmp (BITMAP *bmp, const TEG_MAP* map, int layer, int x, int y, int w, int h, int xview, int yview)
{
}

void teg_partdraw_large (BITMAP *bmp, const TEG_MAP* map, int layer, int x, int y, int w, int h, int xview, int yview)
{
}

void teg_partdraw_largei (BITMAP *bmp, const TEG_MAP* map, int layer, int x, int y, int w, int h, int xview, int yview)
{
}


int teg_pixelw(const TEG_MAP *map)
{
    assert (map);
    if (!map->tilelist) return 0; else return map->w * map->tilelist->tilew;
}

int teg_pixelh(const TEG_MAP *map)
{
    assert (map);
    if (!map->tilelist) return 0; else return map->h * map->tilelist->tileh;
}
